/*
 * DataContainer.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.core;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import ch.uzh.ifi.sparqlml.mining.module.MiningModel;

import com.hp.hpl.jena.graph.Node;
import com.hp.hpl.jena.query.QuerySolution;
import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.sparql.core.Var;

/**
 * Class to create all the necessary data structures (subgraphs) that are needed
 * to do data mining on semantic web data
 * 
 * @author Andr Locher
 */
public class DataContainer {

	/** Constant for datatype string. */
	public static final int STRING = 11;

	/** Constant for datatype date. */
	public static final int DATE = 22;

	/** Constant for datatype integer. */
	public static final int INTEGER = 33;

	/** Constant for datatype float. */
	public static final int FLOAT = 44;

	/** Constant for datatype double. */
	public static final int DOUBLE = 55;

	private MiningModel m;

	public ResultSet results;

	public String rootVar;

	public List<Feature> features;

	public List<String[]> linkList;

	public List<String[]> attrList;

	private Resources objectMap;

	private Subgraphs subgraphs;

	private HashMap<String, String> qsObjects;

	private HashMap<Var, Var> mappings = new HashMap<Var, Var>();

	public DataContainer(ResultSet res, MiningModel model,
			HashMap<Var, Var> mappings) {
		results = res;
		m = model;
		this.mappings = mappings;
		init();
	}

	/**
	 * Initialize necessary pattern information to create ModelData
	 * 
	 */
	public void init() {
		features = m.getModelFeatures();
		linkList = m.getLinkList();
		attrList = m.getAttrList();
		rootVar = features.get(0).getName();
		objectMap = new Resources();
		subgraphs = new Subgraphs(m.getModelName(), features, linkList);
	}

	/**
	 * Return the subgraphs container for the use in a mining module.
	 * 
	 * @return Subgraphs
	 */
	public Subgraphs getSubgraphs() {
		return subgraphs;
	}

	/**
	 * Return the map of objects and their corresponding ID for the use in a
	 * mininig module.
	 * 
	 * @return
	 */
	public Resources getResources() {
		return objectMap;
	}

	/**
	 * Get the variable mapping of the model variable
	 * 
	 * @param in
	 *            The model variable
	 * @return The original resultset variable
	 */
	public String getMappedVar(String in) {
		String out = "";
		Var outVar = null;
		if (mappings == null) {
			out = in;
		} else {
			outVar = mappings.get(Var.alloc(in));
			out = outVar.getName();
		}
		return out;
	}

	/**
	 * Process results and create subgraphs
	 * 
	 */
	public void processResults() {

		Subgraph subgraph = null;

		for (; results.hasNext();) {
			QuerySolution qs = results.nextSolution();
			// only process query solutions with an existing root feature

			if (qs.get(getMappedVar(rootVar)) != null) {

				Node rootNode = qs.get(getMappedVar(rootVar)).asNode();
				qsObjects = new HashMap<String, String>();
				qsObjects.put(rootVar, rootNode.getURI());
				if (!subgraphs.containsSubgraph(rootNode.getURI())) {

					subgraph = new Subgraph(rootNode.getURI());
					subgraph.setDataset(subgraphs);
					String[] rootFeature = { rootNode.getURI() };
					subgraph.addFeatureNode(0, rootFeature);
					subgraphs.addSubgraph(subgraph);

					if (!objectMap.containsObject(rootNode.getURI()))
						objectMap.addObject(rootNode.getURI());
				}

				int featureIndex = 0;

				for (Iterator iter = features.iterator(); iter.hasNext();) {
					// varName, featureType, isPredict, rootVar
					Feature currVar = (Feature) iter.next();
					RDFNode currRDFNode = qs
							.get(getMappedVar(currVar.getName()));

					if (currVar.getName().equals(rootVar)) {
						// do nothing because rootVar is already registered
					} else {

						if (currRDFNode != null) {

							Node currNode = currRDFNode.asNode();

							if (currNode.isURI()) {
								String currValue = currNode.getURI();
								if (!objectMap.containsObject(currValue)) {
									objectMap.addObject(currValue);
								}
								qsObjects.put(currVar.getName(), currValue);
								String[] n = { currValue };
								subgraph.addFeatureNode(featureIndex, n);

							} else if (currNode.isBlank()) {
								String currValue = currNode.getBlankNodeLabel();
								if (!objectMap.containsObject(currValue)) {
									objectMap.addObject(currValue);
								}
								qsObjects.put(currVar.getName(), currValue);
								String[] n = { currValue };
								subgraph.addFeatureNode(featureIndex, n);

							} else if (currNode.isLiteral()) {
								String currValue = currNode
										.getLiteralLexicalForm();

								String resURI = null;
								String predicate = null;

								// get parent resource of literal
								for (Iterator attrIter = attrList.iterator(); attrIter
										.hasNext();) {
									String[] attrItem = (String[]) attrIter
											.next();
									if (attrItem[2].equals(currVar.getName())) {
										String resVar = attrItem[0];
										Node parentRes = qs.get(
												getMappedVar(resVar)).asNode();
										if (parentRes.isURI())
											resURI = parentRes.getURI();
										else if (parentRes.isBlank())
											resURI = parentRes
													.getBlankNodeLabel();
										predicate = attrItem[1];
										break;
									}
								}

								// add datatype to subgraphs header
								String n_datatype = currNode
										.getLiteralDatatypeURI();
								int datatype = -1;
								if (n_datatype != null) {
									n_datatype = n_datatype
											.substring(n_datatype.indexOf('#') + 1);
								} else {
									n_datatype = "";
								}

								if (n_datatype.equals("string"))
									datatype = STRING;
								else if (n_datatype.equals("dateTime"))
									datatype = DATE;
								else if (n_datatype.equals("integer"))
									datatype = INTEGER;
								else if (n_datatype.equals("int"))
									datatype = INTEGER;
								else if (n_datatype.equals("float"))
									datatype = FLOAT;
								else
									datatype = STRING;

								// add datatype
								subgraphs.setDatatype(featureIndex, datatype);

								// add nominal value
								// TODO: REMOVE THIS
								/*
								 * if (datatype.equals("string")) subgraphs
								 * .addNomValue(currVar[0], currValue);
								 */

								// add feature node
								String[] n = { resURI, currValue };
								subgraph.addFeatureNode(featureIndex, n);

							} else {
								// register a corrupt node?
							}

						} else {
							// register a missing value
						}

					}

					featureIndex++;
				}

				// add links to subgraph
				for (int i = 0; i < linkList.size(); i++) {
					String[] linkItem = linkList.get(i);
					String obj1 = qsObjects.get(linkItem[0]);
					String predicate = linkItem[1];
					String obj2 = qsObjects.get(linkItem[2]);
					if (obj1 != null && obj2 != null) {
						String[] lnk = { obj1, predicate, obj2 };
						subgraph.addLinkNode(i, lnk);
					}

				}

			}
		}

	}
}
