/*
 * Feature.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.core;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.List;

/**
 * Class for handling the feature values of a given Feature in a subgraph.
 * 
 * @author Andr Locher
 */
public class Feature {

	/** Constant set for resource features. */
	public static final int RESOURCE = 1;

	/** Constant set for nominal features. */
	public static final int DISCRETE = 2;

	/** Constant set for numeric features. */
	public static final int CONTINUOUS = 3;

	/** Constant for datatype string. */
	public static final int STRING = 11;

	/** Constant for datatype date. */
	public static final int DATE = 22;

	/** Constant for datatype integer. */
	public static final int INTEGER = 33;

	/** Constant for datatype float. */
	public static final int FLOAT = 44;

	/** Constant for datatype double. */
	public static final int DOUBLE = 55;

	/** The feature name */
	private String f_name;

	/** The var name of the parent resource if this is a literal */
	private String f_parent;

	/** The feature type (RESOURCE, DISCRETE or CONTINUOUS) */
	private int f_type;

	/** The datatype of this feature */
	private int f_datatype;

	/** Whether this variable is predictable. 0 if not, 1 if it is. */
	private int f_predict;

	/** The possible values of a discrete/nominal feature. */
	private String[] f_values;

	/** Whether this is the root feature or not */
	private int f_root;

	/** The feature values */
	private List<String[]> nodeList;

	/**
	 * Creates a new DISCRETE feature
	 * 
	 * @param feature
	 *            the name of the feature and its parent
	 * @param data
	 *            the metadata for this feature
	 * @param values
	 *            the nominal values of this feature
	 */
	public Feature(String[] feature, int[] data, String[] values) {
		nodeList = new ArrayList<String[]>();
		f_name = feature[0];
		f_parent = feature[1];
		f_type = data[0];
		f_datatype = data[1];
		f_predict = data[2];
		f_root = data[3];
		f_values = values;
	}

	/**
	 * Creates a new CONTINUOUS feature
	 * 
	 * @param feature
	 *            the name of the feature and its parent
	 * @param data
	 *            the metadata for this feature
	 */
	public Feature(String[] feature, int[] data) {
		nodeList = new ArrayList<String[]>();
		f_name = feature[0];
		f_parent = feature[1];
		f_type = data[0];
		f_datatype = data[1];
		f_predict = data[2];
		f_root = data[3];
	}

	/**
	 * Creates a new RESOURCE feature
	 * 
	 * @param name
	 *            the name of the resource
	 * @param data
	 *            the metadata for this feature
	 */
	public Feature(String name, int[] data) {
		nodeList = new ArrayList<String[]>();
		f_name = name;
		f_type = data[0];
		f_datatype = data[1];
		f_predict = data[2];
		f_root = data[3];
	}

	/**
	 * Creates a new feature by copying an existing feature without it's nodes.
	 * 
	 * @param name
	 *            the name of the resource
	 * @param data
	 *            the metadata for this feature
	 */
	public Feature(Feature f) {
		nodeList = new ArrayList<String[]>();
		f_name = f.f_name;
		f_parent = f.f_parent;
		f_type = f.f_type;
		f_datatype = f.f_datatype;
		f_predict = f.f_predict;
		f_root = f.f_root;
		f_values = f.f_values;
	}

	public Feature(String name, int[] data, String[] values, int i) {
		nodeList = new ArrayList<String[]>();
		f_name = name;
		f_type = data[0];
		f_datatype = data[1];
		f_predict = data[2];
		f_root = data[3];
		f_values = values;
	}

	/**
	 * Adds a new value to this Feature. This adds a value consisting of the
	 * actual literal value and the value of the parent Feature being a
	 * resource.
	 * 
	 * @param n
	 *            the value being a String array
	 */
	public void addNode(String[] n) {
		if (!contains(n)) {
			nodeList.add(n);
		}
	}

	/**
	 * Adds a new value to this Feature.
	 * 
	 * @param n
	 *            the value being a String
	 */
	public void addNode(String n) {
		String[] newVal = { n };
		if (!contains(newVal)) {
			nodeList.add(newVal);
		}
	}

	/**
	 * Gives the number of values of this Feature.
	 * 
	 * @return int being the number of values
	 */
	public int numNodes() {
		return nodeList.size();
	}

	/**
	 * Gets the feature value at the given position.
	 * 
	 * @param index
	 *            representing the position of the value in this feature.
	 * @return string array of the value
	 */
	public String[] getFeatureData(int index) {
		return nodeList.get(index);
	}

	/**
	 * States whether this feature doesn't have any values.
	 * 
	 * @return true if no values present, false otherwise
	 */
	public boolean isMissing() {
		if (nodeList.size() < 1)
			return true;
		return false;
	}

	/**
	 * Checks whether this feature already contains the given value.
	 * 
	 * @param n
	 *            being the value
	 * @return true if value is already present, false otherwise
	 */
	public boolean contains(String[] n) {
		for (int i = 0; i < nodeList.size(); i++) {
			if (Arrays.equals(nodeList.get(i), n)) {
				return true;
			}
		}
		return false;
	}

	/**
	 * Get the name of this feature.
	 * 
	 * @return the name of this feature
	 */
	public String getName() {
		return f_name;
	}

	/**
	 * Returns the possible nominal values of feature.
	 * 
	 * @return String array
	 */
	public String[] getValues() {
		return f_values;
	}

	/**
	 * Returns the content type of this Feature.
	 * 
	 * @return int
	 */
	public int getType() {
		return f_type;
	}

	/**
	 * Returns an integer representing the datatype of this Feature.
	 * 
	 * @return int
	 */
	public int getDatatype() {
		return f_datatype;
	}

	/**
	 * Returns the variable name of the parent feature (resource) if this is a
	 * literal.
	 * 
	 * @return String
	 */
	public String getParent() {
		return f_parent;
	}

	/**
	 * Sets the datatype of this Feature.
	 * 
	 * @param dt
	 */
	public void setDataType(int dt) {
		f_datatype = dt;
	}

	/**
	 * Returns TRUE if the Feature has the content type RESOURCE.
	 * 
	 * @return boolean
	 */
	public boolean isResource() {
		if (f_type == RESOURCE)
			return true;
		return false;
	}

	/**
	 * Returns TRUE if the Feature has the content type DISCRETE.
	 * 
	 * @return boolean
	 */
	public boolean isDiscrete() {
		if (f_type == DISCRETE)
			return true;
		return false;
	}

	/**
	 * Returns TRUE if the Feature has the content type CONTINUOUS.
	 * 
	 * @return boolean
	 */
	public boolean isContinuous() {
		if (f_type == CONTINUOUS)
			return true;
		return false;
	}

	/**
	 * Returns TRUE if the Feature is the target label.
	 * 
	 * @return boolean
	 */
	public boolean isPredict() {
		if (f_predict == 1)
			return true;
		return false;
	}

	/**
	 * Returns TRUE if the Feature is the target object for which we would like
	 * to make a prediction.
	 * 
	 * @return boolean
	 */
	public boolean isRoot() {
		if (f_root == 1)
			return true;
		return false;
	}

	/**
	 * Returns the feature representation as a string printing out all the
	 * values belonging to this feature.
	 */
	public String toString() {
		String output = "";

		for (int i = 0; i < nodeList.size(); i++) {
			String[] curr = nodeList.get(i);
			if (curr.length > 1)
				output += curr[0] + " -> " + curr[1] + "\n";
			else
				output += curr[0] + "\n";
		}
		return output;
	}

}
