/*
 * Subgraph.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.core;

/**
 * Class for handling a subgraph. A subgraph holds all the features and links
 * that have been defined in the subgraphs container.
 * 
 * @author Andr Locher
 */
public class Subgraph {

	/** The name of the root node of this subgraph. */
	protected String root;

	/** The subgraphs container which holds this subgraph */
	protected Subgraphs subgraphs;

	/** The feature container holding all the features for this graph. */
	protected Feature[] featureContainer;

	/** The link container holding all the links for this graph. */
	protected Link[] linkContainer;

	/**
	 * Creates a new subgraph with a root node.
	 * 
	 * @param rootNode
	 *            the name of the root node.
	 */
	public Subgraph(String rootNode) {
		root = rootNode;
	}

	/**
	 * Sets the subgraphs container this subgraph belongs to.
	 * 
	 * @param sgs
	 *            the subgraphs container
	 */
	public void setDataset(Subgraphs sgs) {
		subgraphs = sgs;
		featureContainer = new Feature[subgraphs.numFeatures()];
		for (int i = 0; i < subgraphs.numFeatures(); i++) {
			Feature curr = subgraphs.getFeature(i);
			featureContainer[i] = new Feature(curr);
		}

		linkContainer = new Link[subgraphs.numLinks()];
		for (int i = 0; i < subgraphs.numLinks(); i++) {
			linkContainer[i] = new Link(subgraphs.getLink(i));
		}
	}

	/**
	 * Adds a new feature value to this subgraph at the given position of the
	 * featureContainer.
	 * 
	 * @param index
	 *            representing the position in the featureContainer
	 * @param n
	 *            the value of the feature
	 */
	public void addFeatureNode(int index, String[] n) {
		featureContainer[index].addNode(n);
	}

	/**
	 * Checks whether the feature already exists in the featureContainer.
	 * 
	 * @param index
	 *            representing the position in the featureContainer
	 * @param n
	 *            the value of the feature
	 * @return true if the featureContainer already contains this feature value,
	 *         false otherwise
	 */
	public boolean containsFeatureNode(int index, String[] n) {
		if (featureContainer[index].contains(n))
			return true;
		return false;
	}

	/**
	 * Gets the Feature holding all the values at the given position.
	 * 
	 * @param index
	 *            representing the position in the featureContainer
	 * @return feature of interest
	 */
	public Feature getFeature(int index) {
		return featureContainer[index];
	}

	/**
	 * Gets the Link holding all the possible values at the given position.
	 * 
	 * @param index
	 *            representing the position in the linkContainer
	 * @return link of interest
	 */
	public Link getLink(int index) {
		return linkContainer[index];
	}

	/**
	 * Adds a new link value to this subgraph at the given position of the
	 * linkContainer.
	 * 
	 * @param index
	 *            representing the position in the linkContainer
	 * @param lnk
	 *            the value of the link
	 */
	public void addLinkNode(int index, String[] lnk) {
		linkContainer[index].addLink(lnk);
	}
	
	public String getRoot() {
		return root;
	}

	/**
	 * Returns the subgraph representation as a string printing out the features
	 * and links belonging to this subgraph.
	 */
	public String toString() {
		String output = "";
		output += "--------------------------------" + "\n";
		output += "SUBGRAPH: " + root + "\n";
		output += "--------------------------------" + "\n";
		for (int i = 0; i < featureContainer.length; i++) {
			output += "+++ " + featureContainer[i].getName() + " +++\n";
			output += featureContainer[i].toString();
		}

		output += "LINKS:" + "\n";
		for (int i = 0; i < linkContainer.length; i++) {
			output += "+++ " + linkContainer[i].getName() + " +++\n";
			output += linkContainer[i].toString() + "\n";
		}

		return output;
	}
}
