/*
 * Subgraphs.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.core;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

/**
 * Class for handling an ordered set of subgraphs.
 * 
 * @author Andr Locher
 */
public class Subgraphs {

	/** The name of the subgraphs container */
	protected String s_name;

	/** The subgraphs. */
	protected List<Subgraph> s_data;

	/** Quick access map to subgraphs with root node name */
	protected HashMap<String, Integer> s_map;

	/** The features header for the subgraphs */
	protected List<Feature> s_features;

	/** The links header for the subgraphs */
	protected List<String[]> s_links;

	/** The nominal values of the features */
	protected HashMap<String, String[]> nom_values;

	/** The subgraph counter */
	protected int g_counter;

	/** the class label */
	protected String classLabel;

	/**
	 * Creates and empty set of subgraphs. Uses the given feature and link
	 * information to initialize the header.
	 * 
	 * @param name
	 * @param features
	 * @param links
	 */
	public Subgraphs(String name, List<Feature> features,
			List<String[]> linkList) {
		s_map = new HashMap<String, Integer>();
		s_data = new ArrayList<Subgraph>();
		s_name = name;
		s_links = linkList;
		s_features = features;

		nom_values = new HashMap<String, String[]>();
		for (int i = 0; i < s_features.size(); i++) {
			Feature currFeature = s_features.get(i);
			if ((currFeature.isResource() && currFeature.isPredict())
					|| !currFeature.isResource()) {
				String[] vals = currFeature.getValues();
				nom_values.put(currFeature.getName(), vals);
			}
		}

		// set the classlabel
		for (int j = 0; j < numFeatures(); j++) {
			Feature curr = getFeature(j);
			if (curr.isPredict()) {
				classLabel = curr.getName();
			}
		}
	}

	/**
	 * Adds a subgraph to the ordered set.
	 * 
	 * @param sg
	 */
	public void addSubgraph(Subgraph sg) {
		g_counter++;
		int index = g_counter;
		s_map.put(sg.root, index);
		s_data.add(index - 1, sg);
	}

	/**
	 * Checks whether this set already contains a subgraph with the given root
	 * node name.
	 * 
	 * @param rootURI
	 * @return true if subgraph already exists in this set, false otherwise
	 */
	public boolean containsSubgraph(String rootURI) {
		if (s_map.containsKey(rootURI))
			return true;
		return false;
	}

	/**
	 * Counts the number of features in the subgraphs header.
	 * 
	 * @return an int representing the number of features
	 */
	public int numFeatures() {
		return s_features.size();
	}

	/**
	 * Gets the feature at the given position
	 * 
	 * @param index
	 *            representing the position of the feature
	 * @return string array of the feature definition
	 */
	public Feature getFeature(int index) {
		return s_features.get(index);
	}

	/**
	 * Counts the number of links in the subgraphs header.
	 * 
	 * @return an int representing the number of links
	 */
	public int numLinks() {
		return s_links.size();
	}

	/**
	 * Gets the link at the given position
	 * 
	 * @param index
	 *            representing the position of the link
	 * @return string array of the link definition
	 */
	public String[] getLink(int index) {
		return s_links.get(index);
	}

	/**
	 * Counts the number of subgraphs contained in this set.
	 * 
	 * @return an int representing the number of subgraphs
	 */
	public int numSubgraphs() {
		return s_map.size();
	}

	/**
	 * Gets the subgraph at the given position.
	 * 
	 * @return subgraph at the given position
	 */
	public Subgraph getSubgraph(int index) {
		return s_data.get(index);
	}

	/**
	 * Gets a mapping of root resource and their corresponding subgraph ID
	 * 
	 * @return subgraph at the given position
	 */
	public HashMap<String, Integer> getSubMap() {
		return s_map;
	}

	/**
	 * Sets the datatype of the feature definition in the header.
	 * 
	 * @param index
	 *            representing the position of the feature
	 * @param datatype
	 *            representing a string of the datatype
	 */
	public void setDatatype(int index, int datatype) {
		s_features.get(index).setDataType(datatype);
	}

	/**
	 * Gets the nominal values of a nominal feature
	 * 
	 * @param feature
	 *            of interest
	 * @return array of all possible nominal values
	 */
	public String[] getNomValues(String feature) {
		return nom_values.get(feature);
	}

	/**
	 * Return the class label of the subgraphs in this collection
	 * 
	 * @return the class label
	 */
	public String getClassLabel() {
		return classLabel;
	}

	/**
	 * Returns the subgraphs representation as a string printing out all
	 * subgraphs contained in the set.
	 */
	public String toString() {
		String output = "";
		output += "GRAPH: " + s_name + "\n";
		for (int i = 0; i < s_data.size(); i++) {
			output += s_data.get(i).toString() + "\n";
			output += "\n";
		}
		return output;
	}
}
