/*
 * MiningApp.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.module;

import java.util.HashMap;

import kdl.prox.model2.common.predictions.Predictions;

import com.hp.hpl.jena.query.ResultSet;
import com.hp.hpl.jena.sparql.core.Var;

/**
 * Interface which can be implemented by custom mining modules.
 * 
 * @author Andr Locher
 */
public interface MiningApp {

	/** Path to location of mining models */
	public static String modelsPath = System.getenv("ARQROOT")
			+ System.getProperty("file.separator") + "mining"
			+ System.getProperty("file.separator") + "models"
			+ System.getProperty("file.separator");

	/** Path to location of mining data */
	public static String dataPath = System.getenv("ARQROOT")
			+ System.getProperty("file.separator") + "mining"
			+ System.getProperty("file.separator") + "data"
			+ System.getProperty("file.separator");

	/** MiningApp initialization */
	public void init(MiningModel m);

	/** Initialize (learn) a new model */
	public void learnModel();

	/** Apply a learned model to new data in order to get predictions */
	public void applyModel();

	/**
	 * Data preprocessing for mining module. This method has to transform the
	 * ARQ ResultSet into the generic subgraph structure.
	 * 
	 * @param res
	 *            The resultset of the SPARQL Query
	 * @param mappings
	 *            The mappings to the model variables
	 */
	public void prepareModelData(ResultSet res, HashMap<Var, Var> mappings);

	/**
	 * Get the prediction of a given subgraph after the model has been applied.
	 * 
	 * @param subgID
	 * @return Object Holds the prediction which can be a classLabel or a
	 *         number.
	 */
	public Object getClassPrediction(int subgID);

	/**
	 * Get the probability of the class prediction after the model has been
	 * applied.
	 * 
	 * @param subgID
	 * @return Object Holds the probability of the prediction
	 */
	public Object getClassProbability(int subgID);

	/**
	 * Get the accuracy of the current prediction.
	 * 
	 * @return Object Represents the accuracy value
	 */
	public Object getAccuracy();

	/**
	 * Get the points of the ROC curve
	 * 
	 * @return Object Represents an array of points
	 */
	public Object getROCPoints(String classLabel);

	/**
	 * Get the area under the ROC-curve
	 * 
	 * @return Object Represents the AUC value
	 */
	public Object getAUC(String classLabel);

	/**
	 * Get the collection of subgraphs in the subgraphs container.
	 * 
	 * @return HashMap With the key being the name of the root node and the
	 *         value being the subgraph id
	 */
	public HashMap getSubMap();

	/**
	 * Get the true class of a given subgraph.
	 * 
	 * @return Object Correct class if given
	 */
	public Object getTrueClass(int subgID);

	/**
	 * Get the quared error of a prediction.
	 * 
	 * @return Object Squared error
	 */
	public Object getSquaredError();

}
