/*
 * MiningModel.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.module;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import ch.uzh.ifi.sparqlml.mining.core.Feature;

import com.hp.hpl.jena.rdf.model.Bag;
import com.hp.hpl.jena.rdf.model.Literal;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Statement;
import com.hp.hpl.jena.util.FileManager;

/**
 * Class representing a mining model. Provides methods to access the metadata.
 * 
 * @author Andr Locher
 */
public class MiningModel {

	private Model m;

	private Model settings;

	private static String SMO_PREFIX = "http://www.ifi.uzh.ch/ddis/sparql-ml/";

	private static String miningSettingsFile = System.getenv("ARQROOT")
			+ System.getProperty("file.separator") + "mining"
			+ System.getProperty("file.separator") + "settings.rdf";

	/** Constant for resource features. */
	public static final int RESOURCE = 1;

	/** Constant for nominal features. */
	public static final int DISCRETE = 2;

	/** Constant for numeric features. */
	public static final int CONTINUOUS = 3;

	public MiningModel(Model model) {
		m = model;
		settings = ModelFactory.createDefaultModel();
		InputStream in = FileManager.get().open(miningSettingsFile);
		settings.read(in, "");
	}

	/**
	 * Gets the Jena Model of the current mining model.
	 * 
	 * @return Model
	 */
	public Model getModel() {
		return m;
	}

	public String getModelURI() {
		Resource modelURI = null;
		Iterator i = m.listSubjectsWithProperty(m.createProperty(SMO_PREFIX
				+ "hasModelName"));
		while (i.hasNext()) {
			modelURI = (Resource) i.next();
		}
		return modelURI.getURI();
	}

	public String getModelName() {
		Literal modelName = null;
		Iterator i = m.listObjectsOfProperty(m.createProperty(SMO_PREFIX
				+ "hasModelName"));
		while (i.hasNext()) {
			modelName = (Literal) i.next();
		}
		return modelName.getString();
	}

	public Resource getAlgorithm() {
		Resource alg = null;
		Iterator i = m.listObjectsOfProperty(m.createProperty(SMO_PREFIX
				+ "usesAlgorithm"));
		while (i.hasNext()) {
			alg = (Resource) i.next();
		}
		return alg;
	}

	public String getAlgorithmName() {
		Literal algName = null;
		Iterator i = settings.listObjectsOfProperty(settings
				.createProperty(SMO_PREFIX + "hasAlgorithmName"));
		while (i.hasNext()) {
			algName = (Literal) i.next();
		}
		return algName.getString();
	}

	public String getAlgorithmClass() {
		Literal algorithmClass = null;
		Resource miningAlg = getAlgorithm();
		Iterator i = settings.listObjectsOfProperty(miningAlg, settings
				.createProperty(SMO_PREFIX + "hasClass"));
		while (i.hasNext()) {
			algorithmClass = (Literal) i.next();
		}
		return algorithmClass.getString();
	}

	public List getAlgorithmParams() {
		List<Object> params = new ArrayList<Object>();
		Resource param = null;
		Resource alg = getAlgorithm();

		Iterator i = settings.listObjectsOfProperty(alg, settings
				.createProperty(SMO_PREFIX + "hasParam"));
		while (i.hasNext()) {
			param = (Resource) i.next();
			params.add(getAlgorithmParamData(param));
		}
		return params;
	}

	private Object getAlgorithmParamData(Resource param) {
		String name = null;
		String value = null;
		Iterator iName = settings.listObjectsOfProperty(param, m
				.createProperty(SMO_PREFIX + "hasName"));
		while (iName.hasNext()) {
			Literal n = (Literal) iName.next();
			name = n.getString();
		}
		Iterator iValue = settings.listObjectsOfProperty(param, m
				.createProperty(SMO_PREFIX + "hasValue"));
		while (iValue.hasNext()) {
			Literal n = (Literal) iValue.next();
			value = n.getString();
		}
		String[] paramData = { name, value };
		return paramData;
	}

	public Resource getMiningApp() {
		Resource miningApp = null;
		Resource alg = getAlgorithm();
		Iterator i = settings.listObjectsOfProperty(alg, settings
				.createProperty(SMO_PREFIX + "hasMiningApp"));
		while (i.hasNext()) {
			miningApp = (Resource) i.next();
		}
		return miningApp;
	}

	public String getMiningAppName() {
		Literal miningAppName = null;
		Resource miningApp = getMiningApp();
		Iterator i = settings.listObjectsOfProperty(miningApp, settings
				.createProperty(SMO_PREFIX + "hasAppName"));
		while (i.hasNext()) {
			miningAppName = (Literal) i.next();
		}
		return miningAppName.getString();
	}

	public String getMiningAppClass() {
		Literal miningAppClass = null;
		Resource miningApp = getMiningApp();
		Iterator i = settings.listObjectsOfProperty(miningApp, settings
				.createProperty(SMO_PREFIX + "hasClass"));
		while (i.hasNext()) {
			miningAppClass = (Literal) i.next();
		}
		return miningAppClass.getString();
	}

	/**
	 * Get a custom parameters of a mining module.
	 * 
	 * @param property
	 *            The property name of the custom metadata.
	 * @return Object The object of the given property.
	 */
	public Object getMiningAppObject(String property) {
		Object miningAppObject = null;
		Resource miningApp = getMiningApp();
		Iterator i = settings.listObjectsOfProperty(miningApp, settings
				.createProperty(SMO_PREFIX + property));
		while (i.hasNext()) {
			miningAppObject = i.next();
		}
		return miningAppObject;
	}

	/**
	 * Get the custom paramaters used during the model learning task.
	 * 
	 * @return List Algorithm parameters
	 */
	public List getModelParams() {
		List<String[]> params = new ArrayList<String[]>();
		Resource param = null;
		Iterator i = m.listObjectsOfProperty(m.createProperty(SMO_PREFIX
				+ "hasParam"));
		while (i.hasNext()) {
			param = (Resource) i.next();
			params.add(getModelParamData(param));
		}
		return params;
	}

	private String[] getModelParamData(Resource param) {
		String name = null;
		String value = null;
		Iterator iName = m.listObjectsOfProperty(param, m
				.createProperty(SMO_PREFIX + "hasName"));
		while (iName.hasNext()) {
			Literal n = (Literal) iName.next();
			name = n.getString();
		}
		Iterator iValue = m.listObjectsOfProperty(param, m
				.createProperty(SMO_PREFIX + "hasValue"));
		while (iValue.hasNext()) {
			Literal n = (Literal) iValue.next();
			value = n.getString();
		}
		String[] paramData = { name, value };
		return paramData;
	}

	/**
	 * Get the features used during the model learning task.
	 * 
	 * @return List Features
	 */
	public List<Feature> getModelFeatures() {
		List<Feature> features = new ArrayList<Feature>();
		Resource feature = null;
		Iterator i = m.listObjectsOfProperty(m.createProperty(SMO_PREFIX
				+ "hasFeature"));
		while (i.hasNext()) {
			feature = (Resource) i.next();
			features.add(getModelFeatureData(feature));
		}

		// reorder features: resources (root, ...), literals (..., predict)
		Feature rootFeature = null;
		Feature predictFeature = null;
		for (int x = 0; x < features.size(); x++) {
			Feature curr = features.get(x);
			if (curr.isRoot()) {
				rootFeature = curr;
			} else if (curr.isPredict()) {
				predictFeature = curr;
			} else if (curr.isResource()) {
				features.remove(x);
				features.add(0, curr);
			}
		}

		int indexR = features.indexOf(rootFeature);
		features.remove(indexR);
		features.add(0, rootFeature);

		int indexP = features.indexOf(predictFeature);
		features.remove(indexP);
		features.add(features.size(), predictFeature);

		return features;
	}

	private Feature getModelFeatureData(Resource feature) {

		Feature featureData = null;
		String varName = null;
		int featureType = -1;
		int isPredict = 0;
		int datatype = -1;
		int rootVar = 0;
		String parentVar = null;
		ArrayList<Literal> nominals = new ArrayList<Literal>();
		String[] nomValues;

		Iterator iName = m.listObjectsOfProperty(feature, m
				.createProperty(SMO_PREFIX + "hasVarName"));
		while (iName.hasNext()) {
			Literal n = (Literal) iName.next();
			varName = n.getString();
		}
		Iterator iType = m.listObjectsOfProperty(feature, m
				.createProperty(SMO_PREFIX + "hasFeatureType"));
		while (iType.hasNext()) {
			Literal n = (Literal) iType.next();
			String n_type = n.getString();
			if (n_type.equals("RESOURCE"))
				featureType = 1;
			else if (n_type.equals("DISCRETE"))
				featureType = 2;
			else if (n_type.equals("CONTINUOUS"))
				featureType = 3;
		}
		Iterator iPredict = m.listObjectsOfProperty(feature, m
				.createProperty(SMO_PREFIX + "isPredict"));
		while (iPredict.hasNext()) {
			Literal n = (Literal) iPredict.next();
			isPredict = Integer.valueOf(n.getString());
		}
		Iterator iRoot = m.listObjectsOfProperty(feature, m
				.createProperty(SMO_PREFIX + "isRootVar"));
		while (iRoot.hasNext()) {
			Literal n = (Literal) iRoot.next();
			String n_root = n.getString();
			if (n_root.equals("YES"))
				rootVar = 1;
		}

		// get parent
		if (featureType != RESOURCE) {
			Resource bag = null;
			Iterator i = m.listObjectsOfProperty(feature, m
					.createProperty(SMO_PREFIX + "hasLink"));

			while (i.hasNext()) {
				bag = (Resource) i.next();
				Iterator f = m.listObjectsOfProperty(bag, m
						.createProperty(SMO_PREFIX + "linkFrom"));
				while (f.hasNext()) {
					Resource parentFeature = (Resource) f.next();
					Statement linkStmt = parentFeature.getProperty(m
							.createProperty(SMO_PREFIX + "hasVarName"));
					parentVar = linkStmt.getObject().toString();
				}
			}
		}

		// get nominal values
		if (featureType == DISCRETE
				|| (featureType == RESOURCE && isPredict == 1)) {
			Bag bag = null;
			Iterator i = m.listObjectsOfProperty(feature, m
					.createProperty(SMO_PREFIX + "hasNominalValues"));

			while (i.hasNext()) {
				bag = m.getBag((Resource) i.next());
				Iterator f = bag.iterator();
				while (f.hasNext()) {
					Literal val = (Literal) f.next();
					nominals.add(val);
				}
			}
		}

		// create feature to return
		if (featureType == DISCRETE) {
			String[] var = { varName, parentVar };
			int[] data = { featureType, datatype, isPredict, rootVar };
			nomValues = new String[nominals.size()];
			for (int i = 0; i < nominals.size(); i++) {
				nomValues[i] = nominals.get(i).getString();
			}
			featureData = new Feature(var, data, nomValues);
		} else if (featureType == CONTINUOUS) {
			String[] var = { varName, parentVar };
			int[] data = { featureType, datatype, isPredict, rootVar };
			featureData = new Feature(var, data);
		} else if (featureType == RESOURCE && isPredict == 1) {
			int[] data = { featureType, datatype, isPredict, rootVar };
			nomValues = new String[nominals.size()];
			for (int i = 0; i < nominals.size(); i++) {
				nomValues[i] = nominals.get(i).getString();
			}
			featureData = new Feature(varName, data, nomValues, 1);
		} else if (featureType == RESOURCE) {
			int[] data = { featureType, datatype, isPredict, rootVar };
			featureData = new Feature(varName, data);
		}

		return featureData;
	}

	/**
	 * Get the root feature of the mining model.
	 * 
	 * @return String with the name of the root feature
	 */
	public String getRootFeature() {
		Resource rootRes = null;
		Iterator it = m.listSubjectsWithProperty(m.createProperty(SMO_PREFIX
				+ "isRootVar"));
		while (it.hasNext()) {
			rootRes = (Resource) it.next();
		}
		Statement res = rootRes.getProperty(m.createProperty(SMO_PREFIX
				+ "hasVarName"));
		return res.getObject().toString();
	}

	public String getMiningModelFile() {
		Resource fileName = null;
		Iterator i = m.listObjectsOfProperty(m.createProperty(SMO_PREFIX
				+ "hasModelFile"));
		while (i.hasNext()) {
			fileName = (Resource) i.next();
		}
		return fileName.getLocalName();
	}

	public void setMiningModelFile(String file) {
		Resource modelURI = null;
		Iterator i = m.listSubjectsWithProperty(m.createProperty(SMO_PREFIX
				+ "hasModelName"));
		while (i.hasNext()) {
			modelURI = (Resource) i.next();
		}

		m.add(modelURI, m.createProperty(SMO_PREFIX + "hasModelFile"), m
				.createResource(modelURI.getURI() + "/" + file));
	}

	/**
	 * Gets the link structure of the subgraphs.
	 * 
	 * @return List Link arrays
	 */
	public List<String[]> getLinkList() {
		List<String[]> linkList = new ArrayList<String[]>();
		Resource feature = null;
		Iterator i = m.listSubjectsWithProperty(m.createProperty(SMO_PREFIX
				+ "hasLink"));

		while (i.hasNext()) {
			feature = (Resource) i.next();

			Statement featureTypeStmt = feature.getProperty(m
					.createProperty(SMO_PREFIX + "hasFeatureType"));
			String featureType = featureTypeStmt.getObject().toString();

			if (featureType.equals("RESOURCE")) {
				Iterator f = m.listObjectsOfProperty(feature, m
						.createProperty(SMO_PREFIX + "hasLink"));
				while (f.hasNext()) {
					Resource linkBag = (Resource) f.next();
					Statement linkNameStmt = linkBag.getProperty(m
							.createProperty(SMO_PREFIX + "linkName"));
					Statement linkToStmt = linkBag.getProperty(m
							.createProperty(SMO_PREFIX + "linkTo"));

					String[] linkItem = { feature.getLocalName(),
							linkNameStmt.getObject().toString(),
							linkToStmt.getObject().asNode().getLocalName() };

					linkList.add(linkItem);
				}
			}
		}
		return linkList;
	}

	/**
	 * Gets the attributes/literals of the subgraphs.
	 * 
	 * @return List Attribute arrays
	 */
	public List<String[]> getAttrList() {
		List<String[]> attrList = new ArrayList<String[]>();
		Resource feature = null;
		Iterator i = m.listSubjectsWithProperty(m.createProperty(SMO_PREFIX
				+ "hasLink"));

		while (i.hasNext()) {
			feature = (Resource) i.next();

			Statement featureTypeStmt = feature.getProperty(m
					.createProperty(SMO_PREFIX + "hasFeatureType"));
			String featureType = featureTypeStmt.getObject().toString();

			if (!featureType.equals("RESOURCE")) {
				Iterator f = m.listObjectsOfProperty(feature, m
						.createProperty(SMO_PREFIX + "hasLink"));
				while (f.hasNext()) {
					Resource linkBag = (Resource) f.next();
					Statement linkNameStmt = linkBag.getProperty(m
							.createProperty(SMO_PREFIX + "linkName"));
					Statement linkToStmt = linkBag.getProperty(m
							.createProperty(SMO_PREFIX + "linkFrom"));

					String[] attrItem = {
							linkToStmt.getObject().asNode().getLocalName(),
							linkNameStmt.getObject().toString(),
							feature.getLocalName() };

					attrList.add(attrItem);
				}
			}
		}
		return attrList;
	}

}
