/*
 * MiningModelFactory.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.module;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import ch.uzh.ifi.sparqlml.arq.create.query.QueryML;
import ch.uzh.ifi.sparqlml.mining.core.pattern.GraphPattern;

import com.hp.hpl.jena.rdf.model.Bag;
import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.sparql.expr.Expr;
import com.hp.hpl.jena.sparql.expr.ExprFunction;
import com.hp.hpl.jena.sparql.expr.ExprList;

/**
 * Factory to create a mining model. This factory gets called by the SPARQL-ML
 * CREATE MINING MODEL statement.
 * 
 * @author Andr Locher
 */
public class MiningModelFactory {

	protected static QueryML query;

	private static GraphPattern pattern;

	private static HashSet<String[]> linkList;

	private static HashSet<String[]> attrList;

	private static String SMO_PREFIX = "http://www.ifi.uzh.ch/ddis/sparql-ml/";

	private static String RDF_PREFIX = "http://www.w3.org/1999/02/22-rdf-syntax-ns#";

	protected static HashMap<String, Object[]> nominals;

	private static int hasRoot = 0;

	private static int hasPredict = 0;

	public static MiningModel create(QueryML q) {
		query = q;
		pattern = new GraphPattern(query);
		linkList = pattern.getLinkList();
		attrList = pattern.getAttrList();
		nominals = query.getNominalValues();
		return instantiateModel();
	}

	/**
	 * Returns a new meta mining model built with the information given in the
	 * CREATE MINING MODEL query.
	 * 
	 * @return MiningModel
	 */
	private static MiningModel instantiateModel() {
		Model m = ModelFactory.createDefaultModel();

		m.setNsPrefix("smo", SMO_PREFIX);

		Resource s = m.createResource(query.getMiningModel().getURI());

		m.add(s, m.createProperty(RDF_PREFIX + "type"), m
				.createResource(SMO_PREFIX + "Model"));
		m.add(s, m.createProperty(SMO_PREFIX + "hasModelName"), m
				.createLiteral(query.getMiningModel().getLocalName()));
		m.add(s, m.createProperty(SMO_PREFIX + "usesAlgorithm"), m
				.createResource(query.getMiningAlg().getURI()));

		// get the mining model feature vars
		List<String[]> miningVars = query.getMiningModelVars();
		Iterator vars = miningVars.iterator();
		while (vars.hasNext()) {
			String[] var = (String[]) vars.next();
			addFeature(m, s, var);
		}

		// get the user set algorithm params
		ExprList params = query.getMiningAlgArgs();
		Iterator paramIter = params.iterator();
		while (paramIter.hasNext()) {
			Expr param = (Expr) paramIter.next();
			ExprFunction paramFunc = param.getFunction();
			if (paramFunc != null) {
				addParam(m, s, paramFunc);
			}
		}
		MiningModel mineModel = new MiningModel(m);

		// TODO: Exception handling
		if (hasRoot == 0) {
			System.out.println("Error: no TARGET feature defined.");
			System.exit(0);
		} else if (hasRoot > 1) {
			System.out.println("Error: only one TARGET feature allowed.");
			System.exit(0);
		} else if (hasPredict == 0) {
			System.out.println("Error: no PREDICT feature defined.");
			System.exit(0);
		} else if (hasPredict > 1) {
			System.out.println("Error: only one PREDICT feature allowed.");
			System.exit(0);
		}

		return mineModel;
	}

	/**
	 * Adds the feature information to the meta mining model.
	 * 
	 * @param m
	 *            The current Jena model.
	 * @param s
	 *            The root resource of the meta mining model.
	 * @param var
	 *            The feature which should be added.
	 */
	private static void addFeature(Model m, Resource s, String[] var) {
		String varName = var[0];
		String varFeatureType = var[1];
		int isPredict = Integer.valueOf(var[2]);

		Resource f = m.createResource(query.getMiningModel().getURI() + "#"
				+ varName);

		m.add(s, m.createProperty(SMO_PREFIX + "hasFeature"), f);

		m.add(f, m.createProperty(RDF_PREFIX + "type"), m
				.createResource(SMO_PREFIX + "Feature"));

		// set feature to PREDICT or TARGET
		if (isPredict == 1) {
			m.add(f, m.createProperty(SMO_PREFIX + "isPredict"), String
					.valueOf(isPredict));
			hasPredict++;
		} else if (isPredict == 2 && varFeatureType.equals("RESOURCE")) {
			m.add(f, m.createProperty(SMO_PREFIX + "isRootVar"), "YES");
			hasRoot++;
		}

		m.add(f, m.createProperty(SMO_PREFIX + "hasFeatureType"),
				varFeatureType);
		m.add(f, m.createProperty(SMO_PREFIX + "hasVarName"), varName);

		if (varFeatureType.equals("RESOURCE")) {
			// add linkTo
			for (Iterator linkIter = linkList.iterator(); linkIter.hasNext();) {
				String[] linkItem = (String[]) linkIter.next();
				if (linkItem[0].equals(varName)) {
					Resource link = m.createResource(query.getMiningModel()
							.getURI()
							+ "/link/" + linkItem[1]);

					link.addProperty(m.createProperty(SMO_PREFIX + "linkName"),
							m.createLiteral(linkItem[1]));
					link.addProperty(m.createProperty(SMO_PREFIX + "linkTo"), m
							.createResource(query.getMiningModel().getURI()
									+ "#" + linkItem[2]));

					m.add(link, m.createProperty(RDF_PREFIX + "type"), m
							.createResource(SMO_PREFIX + "Link"));
					m.add(f, m.createProperty(SMO_PREFIX + "hasLink"), link);

				}
			}

			// add nominal values to feature
			Object[] values = nominals.get(varName);
			if (values != null) {
				Bag nomBag = m.createBag();
				for (int i = 0; i < values.length; i++) {
					nomBag.add(m.createLiteral(values[i].toString()));
				}
				m.add(f, m.createProperty(SMO_PREFIX + "hasNominalValues"),
						nomBag);
			}

		} else if (!varFeatureType.equals("RESOURCE")) {
			// add linkFrom
			for (Iterator attrIter = attrList.iterator(); attrIter.hasNext();) {
				String[] attrItem = (String[]) attrIter.next();
				if (attrItem[2].equals(varName)) {
					Resource link = m.createResource(query.getMiningModel()
							.getURI()
							+ "/link/" + attrItem[1]);

					link.addProperty(m.createProperty(SMO_PREFIX + "linkName"),
							m.createLiteral(attrItem[1]));
					link.addProperty(m.createProperty(SMO_PREFIX + "linkFrom"),
							m.createResource(query.getMiningModel().getURI()
									+ "#" + attrItem[0]));

					m.add(link, m.createProperty(RDF_PREFIX + "type"), m
							.createResource(SMO_PREFIX + "Link"));
					m.add(f, m.createProperty(SMO_PREFIX + "hasLink"), link);
				}
			}
			// add nominal values to feature
			Object[] values = nominals.get(varName);
			if (values != null) {
				Bag nomBag = m.createBag();
				for (int i = 0; i < values.length; i++) {
					nomBag.add(m.createLiteral(values[i].toString()));
				}
				m.add(f, m.createProperty(SMO_PREFIX + "hasNominalValues"),
						nomBag);
			}
		}
	}

	/**
	 * Adds algorithm parameters to the meta mining model.
	 * 
	 * @param m
	 *            The current Jena model.
	 * @param s
	 *            The root resource of the meta mining model.
	 * @param paramFunc
	 *            The parameter with its value.
	 */
	private static void addParam(Model m, Resource s, ExprFunction paramFunc) {
		String name = paramFunc.getArg(1).toString().replaceAll("\"", "");
		String value = paramFunc.getArg(2).toString().replaceAll("\"", "");
		m.add(s, m.createProperty(SMO_PREFIX + "hasParam"), m
				.createResource(query.getMiningModel().getURI() + "#" + name));
		Resource f = m.createResource(query.getMiningModel().getURI() + "#"
				+ name);
		m.add(f, m.createProperty(RDF_PREFIX + "type"), m
				.createResource(SMO_PREFIX + "Param"));
		m.add(f, m.createProperty(SMO_PREFIX + "hasValue"), value);
		m.add(f, m.createProperty(SMO_PREFIX + "hasName"), name);
	}
}
