/*
 * ProximityXML.java (10.12.2007)
 * 
 * See LICENSE for more information about licensing and warranties.
 */
package ch.uzh.ifi.sparqlml.mining.module.proximity;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map.Entry;

import kdl.prox.app.ImportTextApp;
import kdl.prox.app.ImportXMLApp;
import kdl.prox.db.DB;
import kdl.prox.util.ImportUtil;
import ch.uzh.ifi.sparqlml.mining.core.Feature;
import ch.uzh.ifi.sparqlml.mining.core.Link;
import ch.uzh.ifi.sparqlml.mining.core.Resources;
import ch.uzh.ifi.sparqlml.mining.core.Subgraph;
import ch.uzh.ifi.sparqlml.mining.core.Subgraphs;
import ch.uzh.ifi.sparqlml.mining.module.BaseWriter;

/**
 * Class that handles the export of subgraphs to Proximity XML format.
 * 
 * Proximity: http://kdl.cs.umass.edu/proximity/
 * 
 * @author Andr Locher
 */
public class ProximityXML extends BaseWriter {

	/** Constant for datatype string. */
	public static final int STRING = 11;

	/** Constant for datatype date. */
	public static final int DATE = 22;

	/** Constant for datatype integer. */
	public static final int INTEGER = 33;

	/** Constant for datatype float. */
	public static final int FLOAT = 44;

	/** Constant for datatype double. */
	public static final int DOUBLE = 55;

	public Subgraphs s;

	public HashMap<String, Integer> resObjects;

	public ProximityXML(Subgraphs subgraphs, String model, Resources o) {
		super(model, "xml");
		s = subgraphs;
		resObjects = o.objects;
	}

	public void write() {
		writeHeader();
		writeObjects();
		writeLinks();
		writeAttributes();
		writeContainer();
		writerFooter();

		importToProximity();
	}

	private void importToProximity() {

		ImportXMLApp im = new ImportXMLApp();
		File tempDir = null;
		try {
			tempDir = File.createTempFile("prox-xml-import", "");
			if (!tempDir.delete())
				throw new IOException();
			if (!tempDir.mkdir())
				throw new IOException();
			final String tempDirPath = tempDir.getPath();

			// convert XML to Text
			im.convertXMLToText(new File(dataPath + model + ".xml"), true,
					tempDirPath);

			// load text and convert to prox
			new ImportTextApp(tempDirPath);
			DB.commit();

		} catch (Exception e) {
		} finally {
			// TODO: workaround, otherwise windows doesn't delete files
			System.gc();
			deleteDirectory(tempDir);
		}

	}

	static public boolean deleteDirectory(File path) {
		if (path.exists()) {
			File[] files = path.listFiles();
			for (int i = 0; i < files.length; i++) {
				if (files[i].isDirectory()) {
					deleteDirectory(files[i]);
				} else {
					files[i].delete();
				}
			}
		}
		return (path.delete());
	}

	private void writeHeader() {
		writer.print("<?xml version=\"1.0\" encoding=\"ISO-8859-1\"?>\n");
		writer.print("<!DOCTYPE PROX3DB SYSTEM \"" + "dtd"
				+ System.getProperty("file.separator") + "prox3db.dtd\">\n");
		writer.print("<PROX3DB>\n");
	}

	private void writerFooter() {
		writer.println("</PROX3DB>");
		writer.close();
	}

	private void writeObjects() {
		writer.print(space + "<OBJECTS>\n");
		Iterator iter = resObjects.entrySet().iterator();
		while (iter.hasNext()) {
			Entry e = (Entry) iter.next();
			Integer objID = (Integer) e.getValue();

			// WRITE OBJECT
			writer.print(space + space + "<OBJECT ID=\"" + objID.intValue()
					+ "\"/>\n");
		}
		writer.print(space + "</OBJECTS>\n");
	}

	private void writeLinks() {
		writer.print(space + "<LINKS>\n");

		int linkCounter = 1;

		for (int i = 0; i < s.numLinks(); i++) {

			for (int z = 0; z < s.numSubgraphs(); z++) {
				Subgraph currSub = s.getSubgraph(z);
				Link links = currSub.getLink(i);

				if (links.numLinks() > 0) {

					for (int x = 0; x < links.numLinks(); x++) {
						String[] link = links.getLinkData(x);

						int obj1 = resObjects.get(link[0]);
						int obj2 = resObjects.get(link[2]);
						// WRITE LINK
						writer.print(space + space + "<LINK ID=\""
								+ linkCounter + "\" O1-ID=\"" + obj1
								+ "\" O2-ID=\"" + obj2 + "\"/>\n");
						linkCounter++;
					}
				}
			}

		}
		writer.print(space + "</LINKS>\n");
	}

	private void writeAttributes() {
		writer.print(space + "<ATTRIBUTES>\n");

		for (int y = 0; y < s.numFeatures(); y++) {
			Feature attr = s.getFeature(y);
			ArrayList<Integer> resAttributes = new ArrayList<Integer>();

			if (!attr.isResource()) {

				String datatype = "str";
				int dtype = attr.getDatatype();
				switch (dtype) {
				case STRING:
					datatype = "str";
					break;
				case DATE:
					datatype = "date";
					break;
				case INTEGER:
					datatype = "int";
					break;
				case FLOAT:
					datatype = "flt";
					break;
				case DOUBLE:
					datatype = "dbl";
					break;
				default:
					datatype = "str";
				}

				writer.print(space + space + "<ATTRIBUTE NAME=\""
						+ attr.getName() + "\" ITEM-TYPE=\"" + "O"
						+ "\" DATA-TYPE=\"" + datatype + "\">\n");

				// walk through all subgraphs
				for (int z = 0; z < s.numSubgraphs(); z++) {
					Subgraph currSub = s.getSubgraph(z);
					Feature data = currSub.getFeature(y);

					// walk through all values of this subgraph
					for (int i = 0; i < data.numNodes(); i++) {

						int objID = resObjects.get(data.getFeatureData(i)[0]);
						if (!resAttributes.contains(objID)) {
							resAttributes.add(objID);
							writer.print(space + space + space
									+ "<ATTR-VALUE ITEM-ID=\"" + objID
									+ "\">\n");
							writer.print(space + space + space + space
									+ "<COL-VALUE>" + data.getFeatureData(i)[1]
									+ "</COL-VALUE>\n");
							writer.print(space + space + space
									+ "</ATTR-VALUE>\n");
						}
					}

				}

				writer.print(space + space + "</ATTRIBUTE>\n");
			} else if (!attr.isRoot() && attr.isPredict()) {

				// handling of resource to be predicted
				String datatype = "str";
				writer.print(space + space + "<ATTRIBUTE NAME=\""
						+ attr.getName() + "\" ITEM-TYPE=\"" + "O"
						+ "\" DATA-TYPE=\"" + datatype + "\">\n");

				for (int z = 0; z < s.numSubgraphs(); z++) {
					Subgraph currSub = s.getSubgraph(z);
					Feature data = currSub.getFeature(y);

					// walk through all values of this subgraph
					if (attr.isPredict()) {
						int objID = resObjects.get(currSub.getRoot());
						writer.print(space + space + space
								+ "<ATTR-VALUE ITEM-ID=\"" + objID + "\">\n");
						if (data.isMissing()) {

						} else {
							writer.print(space + space + space + space
									+ "<COL-VALUE>" + data.getFeatureData(0)[0]
									+ "</COL-VALUE>\n");
						}
						writer.print(space + space + space + "</ATTR-VALUE>\n");
					}
				}
				writer.print(space + space + "</ATTRIBUTE>\n");

				/*
				 * // handling of resource to be predicted (existence) String
				 * datatype = "str"; writer.print(space + space + "<ATTRIBUTE
				 * NAME=\"" + attr.getName() + "\" ITEM-TYPE=\"" + "O" + "\"
				 * DATA-TYPE=\"" + datatype + "\">\n");
				 * 
				 * for (int z = 0; z < s.numSubgraphs(); z++) { Subgraph currSub =
				 * s.getSubgraph(z); Feature data = currSub.getFeature(y); //
				 * walk through all values of this subgraph if
				 * (attr.isPredict()) { int objID =
				 * resObjects.get(currSub.getRoot()); writer.print(space + space +
				 * space + "<ATTR-VALUE ITEM-ID=\"" + objID + "\">\n"); if
				 * (data.isMissing()) { writer.print(space + space + space +
				 * space + "<COL-VALUE>" + "NO" + "</COL-VALUE>\n"); } else {
				 * writer.print(space + space + space + space + "<COL-VALUE>" +
				 * "YES" + "</COL-VALUE>\n"); } writer.print(space + space +
				 * space + "</ATTR-VALUE>\n"); } } writer.print(space + space + "</ATTRIBUTE>\n");
				 */
			}

		}

		// WRITE LINK ATTRIBUTE
		int linkCounter = 1;
		boolean classLabel = false;

		for (int j = 0; j < s.numLinks(); j++) {

			// don't add type attribute for class resource
			if (s.getLink(j)[2].equals(s.getClassLabel())) {
				classLabel = true;
			} else {
				classLabel = false;
			}

			if (!classLabel) {
				writer.print(space + space + "<ATTRIBUTE NAME=\""
						+ s.getLink(j)[2] + "\" ITEM-TYPE=\"" + "L"
						+ "\" DATA-TYPE=\"" + "str" + "\">\n");
			}

			for (int z = 0; z < s.numSubgraphs(); z++) {
				Subgraph currSub = s.getSubgraph(z);
				Link links = currSub.getLink(j);

				if (links.numLinks() > 0) {

					for (int x = 0; x < links.numLinks(); x++) {
						String[] link = links.getLinkData(x);

						if (!classLabel) {
							writer.print(space + space + space
									+ "<ATTR-VALUE ITEM-ID=\"" + linkCounter
									+ "\">\n");
							writer.print(space + space + space + space
									+ "<COL-VALUE>" + link[2]
									+ "</COL-VALUE>\n");
							writer.print(space + space + space
									+ "</ATTR-VALUE>\n");
							linkCounter++;
						} else {
							linkCounter++;
						}
					}
				}

			}
			if (!classLabel) {
				writer.print(space + space + "</ATTRIBUTE>\n");

			}

		}
		writer.print(space + "</ATTRIBUTES>\n");
	}

	private void writeContainer() {
		writer.print(space + "<CONTAINERS>\n");
		writer.print(space + space + "<CONTAINER NAME=\"" + model
				+ "\" ID=\"0\">\n");
		writer.print(space + space + space + "<SUBG-ITEMS>\n");

		int linkCounter = 1;
		boolean firstRun = true;

		// add container for propositional data
		if (s.numLinks() < 1) {
			for (int z = 0; z < s.numSubgraphs(); z++) {
				Subgraph currSub = s.getSubgraph(z);

				// ADD OBJECTS
				if (firstRun) {
					for (int i = 0; i < s.numFeatures(); i++) {
						Feature feat = currSub.getFeature(i);

						if (feat.isResource()) {
							for (int x = 0; x < feat.numNodes(); x++) {
								String[] curr = feat.getFeatureData(x);
								int objID = resObjects.get(curr[0]);
								writer.print(space + space + space + space
										+ "<ITEM SUBG-ID=\"" + (z + 1)
										+ "\" ITEM-ID=\"" + objID
										+ "\" ITEM-TYPE=\"" + "O"
										+ "\" NAME=\"" + feat.getName()
										+ "\"/>\n");
							}
						}
					}
				}
			}
		}

		// add container for relational data
		for (int j = 0; j < s.numLinks(); j++) {

			for (int z = 0; z < s.numSubgraphs(); z++) {
				Subgraph currSub = s.getSubgraph(z);
				Link links = currSub.getLink(j);

				// ADD OBJECTS
				if (firstRun) {
					for (int i = 0; i < s.numFeatures(); i++) {
						Feature feat = currSub.getFeature(i);

						if (feat.isResource()) {
							for (int x = 0; x < feat.numNodes(); x++) {
								String[] curr = feat.getFeatureData(x);
								int objID = resObjects.get(curr[0]);
								writer.print(space + space + space + space
										+ "<ITEM SUBG-ID=\"" + (z + 1)
										+ "\" ITEM-ID=\"" + objID
										+ "\" ITEM-TYPE=\"" + "O"
										+ "\" NAME=\"" + feat.getName()
										+ "\"/>\n");
							}
						}
					}
				}

				// ADD LINKS

				if (links.numLinks() > 0) {

					for (int x = 0; x < links.numLinks(); x++) {
						String[] link = links.getLinkData(x);

						String name = "link_" + link[1];

						writer.print(space + space + space + space
								+ "<ITEM SUBG-ID=\"" + (z + 1)
								+ "\" ITEM-ID=\"" + linkCounter
								+ "\" ITEM-TYPE=\"" + "L" + "\" NAME=\"" + name
								+ "\"/>\n");
						linkCounter++;
					}
				}

			}
			firstRun = false;

		}
		writer.print(space + space + space + "</SUBG-ITEMS>\n");
		writer.print(space + space + "</CONTAINER>\n");
		writer.print(space + "</CONTAINERS>\n");
	}

}
